#import libs

import json
import random

import torch
import torch.nn as nn

import numpy as np
import nltk
from nltk.stem.porter import PorterStemmer

# make instances

stemmer = PorterStemmer()

#model

class eureka_intent_model(nn.Module):
    def __init__(self, input_size, hidden_size, num_classes):
        super(eureka_intent_model, self).__init__()
        self.l1 = nn.Linear(input_size, hidden_size)
        self.l2 = nn.Linear(hidden_size, hidden_size)
        self.l3 = nn.Linear(hidden_size, num_classes)
        self.relu = nn.ReLU()
    
    def forward(self, x):
        out = self.l1(x)
        out = self.relu(out)
        out = self.l2(out)
        out = self.relu(out)
        out = self.l3(out)
        return out

# get tokenizers, stems, etc

def word_stem(word):
    return stemmer.stem(word.lower())

def tokenize(sentence):
    return nltk.word_tokenize(sentence)

def bag_of_words(tokenized_sentence, words):
    sentence_words = []
    for word in tokenized_sentence:
        stemmed_word = word_stem(word)
        sentence_words.append(stemmed_word)
    
    bag = np.zeros(len(words), dtype=np.float32)
    for idx, w in enumerate(words):
        if w in sentence_words:
            bag[idx] = 1
    return bag

def load_data(datafile):
    data = torch.load(datafile)

    return data["input_size"], data["hidden_size"], data["output_size"], data['all_words'], data['tags'], data['model_state']

def get_device():
    if torch.cuda.is_available():
        device = torch.device('cuda')
        print("Cuda.")
    else:
        print("CPU.")
        device = torch.device('cpu')
    
    return device

def get_intents(path):
    with open(path, "r") as file:
        return json.load(file)

class sheil_ai:
    def __init__(self, datapath):
        self.input_size, self.hidden_size, self.output_size, self.all_words, self.tags, self.model_state = load_data(datapath)
        self.device = get_device()
        self.intents, self.threshold = None, None
        self.model = eureka_intent_model(self.input_size, self.hidden_size, self.output_size).to(self.device)
        self.model.load_state_dict(self.model_state)
        self.model.eval()

    def set_intents(self, path):
        self.intents = get_intents(path)
    
    def set_data(self, path):
        self.datafile = load_data(path)
    
    def set_threshold(self, value):
        self.threshold = value
    
    def word_stem(self, word):
        return stemmer.stem(word.lower())

    def tokenize(self, sentence):
        return nltk.word_tokenize(sentence)
    
    def bag_of_words(self, tokenized_sentence, words):
        sentence_words = []
        for word in tokenized_sentence:
            stemmed_word = self.word_stem(word)
            sentence_words.append(stemmed_word)
        
        bag = np.zeros(len(words), dtype=np.float32)
        for idx, w in enumerate(words):
            if w in sentence_words:
                bag[idx] = 1
        return bag
    
    def classify(self, sentence):
        # check if self.data and self.intents exist. raise error if not.
        InputTokens = self.tokenize(sentence)
        X = self.bag_of_words(sentence, self.all_words)
        X = X.reshape(1, X.shape[0])
        X = torch.from_numpy(X).to(self.device)

        output = self.model(X)

        ph, predicted = torch.max(output, dim=1)

        tag = self.tags[predicted.item()]

        probabilities = torch.softmax(output, dim=1)
        probability = probabilities[0][predicted.item()]

        for intent in self.intents['intents']:
            if tag == intent["tag"]:
                return intent

if __name__ == "__main__":
    Sheila = sheil_ai("./data.pth")
    Sheila.set_intents("./intents.json")
    Sheila.set_threshold(0.75)

    UserInput = input("You: ")
    intent_class = Sheila.classify(UserInput)
    print(intent_class)


import numpy as np
import random
import json

import torch
import torch.nn as nn
from torch.utils.data import Dataset, DataLoader

from .sheilai import *

with open('intents.json', 'r') as f:
    intents = json.load(f)

all_words = []
tags = []
xy = []
# loop through each sentence in our intents patterns
for intent in intents['intents']:
    tag = intent['tag']
    # add to tag list
    tags.append(tag)
    for pattern in intent['patterns']:
        # tokenize each word in the sentence
        w = tokenize(pattern)
        # add to our words list
        all_words.extend(w)
        # add to xy pair
        xy.append((w, tag))

# word_stem and lower each word
ignore_words = ['?', '.', '!']
all_words = [word_stem(w) for w in all_words if w not in ignore_words]
# remove duplicates and sort
all_words = sorted(set(all_words))
tags = sorted(set(tags))

print(len(xy), "patterns")
print(len(tags), "tags:", tags)
print(len(all_words), "unique word_stemmed words:", all_words)

# create training data
X_train = []
y_train = []
for (pattern_sentence, tag) in xy:
    # X: bag of words for each pattern_sentence
    bag = bag_of_words(pattern_sentence, all_words)
    X_train.append(bag)
    # y: PyTorch CrossEntropyLoss needs only class labels, not one-hot
    label = tags.index(tag)
    y_train.append(label)

X_train = np.array(X_train)
y_train = np.array(y_train)

# Hyper-parameters 
num_epochs = 1000
batch_size = 8
learning_rate = 0.001
input_size = len(X_train[0])
hidden_size = 8
output_size = len(tags)
print(input_size, output_size)

class ChatDataset(Dataset):

    def __init__(self):
        self.n_samples = len(X_train)
        self.x_data = X_train
        self.y_data = y_train

    # support indexing such that dataset[i] can be used to get i-th sample
    def __getitem__(self, index):
        return self.x_data[index], self.y_data[index]

    # we can call len(dataset) to return the size
    def __len__(self):
        return self.n_samples

dataset = ChatDataset()
train_loader = DataLoader(dataset=dataset,
                          batch_size=batch_size,
                          shuffle=True,
                          num_workers=0)

device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

model = eureka_intent_model(input_size, hidden_size, output_size).to(device)

# Loss and optimizer
criterion = nn.CrossEntropyLoss()
optimizer = torch.optim.Adam(model.parameters(), lr=learning_rate)

# Train the model
for epoch in range(num_epochs):
    for (words, labels) in train_loader:
        words = words.to(device)
        labels = labels.to(dtype=torch.long).to(device)
        
        # Forward pass
        outputs = model(words)
        # if y would be one-hot, we must apply
        # labels = torch.max(labels, 1)[1]
        loss = criterion(outputs, labels)
        
        # Backward and optimize
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()
        
    if (epoch+1) % 100 == 0:
        print (f'Epoch [{epoch+1}/{num_epochs}], Loss: {loss.item():.4f}')


print(f'final loss: {loss.item():.4f}')

data = {
"model_state": model.state_dict(),
"input_size": input_size,
"hidden_size": hidden_size,
"output_size": output_size,
"all_words": all_words,
"tags": tags
}

FILE = "data.pth"
torch.save(data, FILE)

print(f'training complete. file saved to {FILE}')
